#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "spinsubs.h"

#define MAX_LINE_SIZE 400

#define LEVEL_TYPE_CASE         1
#define LEVEL_TYPE_CASE_ITEM    2
#define LEVEL_TYPE_REPEAT       3
#define LEVEL_TYPE_RETURN       4
#define LEVEL_TYPE_REPEAT_FROM  5

#define MODE_CON 1
#define MODE_DAT 2
#define MODE_VAR 3
#define MODE_PUB 4
#define MODE_OBJ 5

#define REPEAT_FROM  1
#define REPEAT_STEP  2
#define REPEAT_COUNT 3
#define REPEAT_FOR   4

void PrintTokens(char **tokens, int tokenum, FILE *outfile);
void ProcessPrototypes(FILE *infile, int unicode, char *pub_pri, FILE *outfile);

#undef  DAT_STRUCT

int indent;
int level = 0;
int levind[10];
int levtype[10];
char *levstr[10];

char combuf[20000];
char *comptr = combuf;

extern int numsym;
extern symbolT symbol[1000];

typedef struct DatStateS {
    int col;
    int mode;
    int notfirst;
} DatStateT;

static char *types[3] = {"byte", "word", "long"};

void AddComment(char *str)
{
    if (*str == 0) str = " ";
    strcpy(comptr, str);
    comptr += strlen(str) + 1;
    *comptr = 0;
}

void PrintComments(FILE *outfile)
{
    char *ptr = combuf;
    while (*ptr)
    {
        fprintf(outfile, "%s\n", ptr);
        ptr += strlen(ptr) + 1;
    }
    comptr = combuf;
    *comptr = 0;
}

extern char *bigptr;

static char *pubptr = "result";

typedef struct OpsListS {
    char *spinstr;
    char *cstr;
    char flag;
    char prec;
} OpsListT;

OpsListT OpsList[] = {
    "===",   "==",       6, 12,
    "<>=",   "!=",       6, 12,
    "<<=",   0,          0, 12,
    ">>=",   "SHR",      5, 12,
    "~>=",   ">>=",      1, 12,
    "<#=",   "LIM_MAX",  5, 12,
    "#>=",   "LIM_MIN",  5, 12,
    "|<=",   "DEC",      5, 12,
    ">|=",   "ENC",      5, 12,
    "//=",   "%=",       1, 12,
    ":=",    "=",        1, 12,
    "<=",    "<=",       1, 12,
    ">=",    ">=",       1, 12,
    "/=",    0,          0, 12,
    "-=",    0,          0, 12,
    "+=",    0,          0, 12,
    "|=",    0,          0, 12,
    "&=",    0,          0, 12,
    "^=",    0,          0, 12,
    "!=",    "~=",       1, 12,
    "or",    "||",       1, 11,
    "and",   "&&",       1, 10,
    "not",   "!",        1,  9,
    "=>",    ">=",       1,  8,
    "=<",    "<=",       1,  8,
    "==",    "==",       1,  8,
    "<>",    "!=",       1,  8,
    "<#",    "LIM_MAX",  2,  7,
    "#>",    "LIM_MIN",  2,  7,
    "//",    "%",        1,  5,
    "/",     0,          0,  5,
    "**",    "MULHIGH",  2,  5,
    "^",     0,          0,  4,
    "->",    "ROR",      2,  2,
    "<<",    "<<",       1,  2,
    ">>",    "SHR",      2,  2,
    "<-",    "ROL",      2,  2,
    "~>",    "SAR",      2,  2,
    "><",    "REV",      2,  2,
    "^^",    "SQRT",     3,  1,
    "||",    "ABS",      3,  1,
    "|<",    "DEC",      3,  1,
    ">|",    "ENC",      3,  1,
    "@@",    0,          0,  0,
    "--",    0,          0,  0,
    "++",    0,          0,  0,
    "~~",    "SEXB",     3,  0,
    "<",     0,          0,  8,
    ">",     0,          0,  8,
    "-",     0,          0,  6,
    "+",     0,          0,  6,
    "*",     0,          0,  5,
    "|",     0,          0,  4,
    "&",     0,          0,  3,
    "!",     "~",        1,  1,
    "@",     0,          4,  0,
    "~",     "SEXW",     3,  0,
    "quit",  "break",    1, 99,
    "next",  "continue", 1, 99,
    "\\",    0,          4, 99,
    "[",     0,          4, 99,
    "]",     0,          4, 99,
    "(",     0,          4, 99,
    ")",     0,          4, 99,
    0,       0,          0,  0};


char *typelist[] = {"byte", "word", "long", 0};

void usage(void)
{
    printf("usage: file[.ext]\n");
    exit(1);
}

int StrCompNoCase(char *str1, char *str2)
{
    while (*str1)
    {
        if (toupper(*str1) != toupper(*str2)) return 1;
        str1++;
        str2++;
    }
    return *str2;
}

void PrintBlanks(int num, FILE *outfile)
{
    while (num-- > 0) fprintf(outfile, " ");
}

char *FindChar(char *str, int val)
{
    while (*str)
    {
	if (*str == val) break;
	str++;
    }
    return str;
}

char *FindChars(char *str, char *str1)
{
    char *str2;

    while (*str)
    {
	str2 = str1;
	while (*str2)
	{
	    if (*str == *str2++) return str;
	}
	str++;
    }
    return str;
}

char *SkipChars(char *str, char *str1)
{
    char *str2;

    while (*str)
    {
	str2 = str1;
	while (*str2)
	{
	    if (*str == *str2) break;
	    str2++;
	}
	if (*str2 == 0) break;
	str++;
    }
    return str;
}

void RemoveCRLF(char *str)
{
    int len = strlen(str);

    str += len - 1;

    while (len)
    {
	if (*str != 10 && *str != 13) break;
	*str-- = 0;
	len--;
    }
}

int GetIndent(char *str)
{
    int indent = 0;

    while (*str)
    {
	if (*str == ' ')
	    indent++;
	else if (*str == '\t')
	    indent = (indent + 8) & 0xfffffff8;
	else
	    break;
	str++;
    }
    return indent;
}

int SearchList(char *str, char **list)
{
    int i = 0;

    while (list[i])
    {
	if (strncmp(str, list[i], strlen(list[i])) == 0)
	    return i;
	i++;
    }
    return -1;
}

int SearchOpsList(char *str, OpsListT *opslist)
{
    int len;
    int i = 0;

    while (opslist[i].spinstr)
    {
        len = strlen(opslist[i].spinstr);
	if (strncmp(str, opslist[i].spinstr, len) == 0)
        {
            int val = str[len];
            if (isalpha(opslist[i].spinstr[0]) && (val == '_' ||
                isalpha(val) || isdigit(val))) return -1;
	    return i;
        }
	i++;
    }
    return -1;
}

int IsUnaryOp(char *str)
{
    static char *UnaryOps = "+-~!&*";
    char *ptr = FindChar(UnaryOps, *str);

    if (*ptr == 0) return 0;
    if (str[1] == 0) return 1;
    if (str[2] != 0) return 0;
    if (str[0] == '+' && str[1] == '+') return 1;
    if (str[0] == '-' && str[1] == '-') return 1;
    return 0;
}

int IsVarChar(int val)
{
    if (isdigit(val) || isalpha(val) || val == '_') return 1;
    return 0;
}

int LocateNextVar(char **tokens, int tokenum, int position)
{
    int i;
    int parencount = 0;

    //printf("LocateNextVar: ");
    //PrintTokens(tokens, tokenum, stdout);
    for (i = position; i < tokenum; i++)
    {
        //printf("LocateNextVar: %s\n", tokens[i]);
	if (tokens[i][0] == ')') parencount++;
	else if (tokens[i][0] == '(') parencount--;
	else if (tokens[i][0] == ']') parencount++;
	else if (tokens[i][0] == '[') parencount--;
        if (parencount > 0) parencount = 0;
        if (parencount) continue;
        if (tokens[i][0] == ')' || tokens[i][0] == ']') continue;
        if (strcmp(tokens[i], "++") == 0) continue;
        if (strcmp(tokens[i], "--") == 0) continue;
        if (IsVarChar(tokens[i][0])) continue;
        if (tokens[i][0] == '$' || tokens[i][0] == '%') continue;
        if (IsUnaryOp(tokens[i]) && i == position) continue;
        break;
    }
    //printf("LocateNextVar: return %s\n", tokens[i-1]);
    return i;
}

int LocatePrevVar(char **tokens, int tokenum, int position)
{
    int i;
    int parencount = 0;

    //printf("LocatePrevVar: ");
    //PrintTokens(tokens, position+1, stdout);
    for (i = position; i > 0; i--)
    {
        //printf("LocatePrevVar: %s\n", tokens[i]);
	if (tokens[i][0] == ')') parencount++;
	else if (tokens[i][0] == '(') parencount--;
	else if (tokens[i][0] == ']') parencount++;
	else if (tokens[i][0] == '[') parencount--;
	if (parencount == 0)
        {
            if (IsVarChar(tokens[i][0]) &&
                (tokens[i-1][0] == '(' || tokens[i-1][0] == '['))
            {
                parencount++;
                continue;
            }
            else if (tokens[i][0] == '(' || IsVarChar(tokens[i][0]))
            {
                if (tokens[i-1][0] == ')') continue;
                if (tokens[i][0] == '(' && IsVarChar(tokens[i-1][0])) continue;
                if (IsUnaryOp(tokens[i-1]))
                {
                    if (tokens[i-2][0] == ')' || tokens[i-2][0] == ']' ||
                        IsVarChar(tokens[i-2][0])) break;
                    else
                        continue;
                }
                break;
            }
            else if (tokens[i][0] == '[') continue;
            else
                break;
        }
    }
    //printf("LocatePrevVar: return %s\n", tokens[i]);
    return i;
}

int InsertToken(char **tokens, int tokenum, int position, char *token)
{
    int i;

    for (i = tokenum; i > position; i--)
        tokens[i] = tokens[i-1];

    tokens[position] = token;

    return tokenum + 1;
}

int RemoveToken(char **tokens, int tokenum, int position)
{
    int i;

    for (i = position; i < tokenum; i++)
        tokens[i] = tokens[i+1];

	return tokenum - 1;
}

int tokenize(char **tokens, char *str, char *tokstr)
{
    int j;
    int len;
    char *str1;
    char *str2;
    int num = 0;

    while (*str)
    {
        str = SkipChars(str, " \t");
	if (*str == 0) break;
        if (*str == '"')
	{
	    str1 = FindChar(str + 1, '"');
	}
	else if (*str == '\'')
	{
	    len = strlen(str);
	    str1 = str + len - 1;
	}
	else if (*str == ',')
	{
	    str1 = str;
	}
	else if ((j = SearchOpsList(str, OpsList)) >= 0)
	{
	    len = strlen(OpsList[j].spinstr);
	    str1 = str + len - 1;
	}
	else
	{
	    str1 = FindChars(str + 1, " \t,|$()<>[]:=+-/*@~\\%");
	    str2 = FindChar(str + 1, '.');
	    if (str2 < str1 && strncmp(str2, ".byte", 5) == 0) str1 = str2;
	    if (str2 < str1 && strncmp(str2, ".word", 5) == 0) str1 = str2;
	    str1--;
	}
	len = str1 - str + 1;
	tokens[num++] = tokstr;
	memcpy(tokstr, str, len);
	tokstr[len] = 0;
	tokstr += len + 1;
	str += len;
    }
    return num;
}

int IsConstant(char *str)
{
    int j;
    int val = *str;

    if (val == '$') return 1;

    if (val == '%') return 1;

    if (val >= '0' && val <= '9') return 1;

    if (val == '"') return 1;

    j = FindSymbol(str);

    if (j < 0) return 0;

    if (GetSymbolType(j) == TYPE_CONSTANT) return 1;

    return 0;
}

char *ProcessBinaryConstant(char *str)
{
    int val = 0;
    int num = 0;
    char *ptr = bigptr;
    char *hexchar = "0123456789abcdef";

    while (*str)
    {
	if (*str == '0' || *str == '1')
	{
	    val = (val << 1) | ((*str) & 1);
	    num++;
	}
	str++;
    }
    num = (num+3) & 0xfffffffc;

    *bigptr++ = '0';
    *bigptr++ = 'x';
    while (num > 0)
    {
	num -= 4;
	*bigptr++ = hexchar[(val >> num) & 15];
    }
    *bigptr++ = 0;
    return ptr;
}

int PrintNumber(char *str, FILE *outfile)
{
    int num = 0;
    while (*str)
    {
        if (*str != '_') fprintf(outfile, "%c", *str);
        str++;
        num++;
    }
    return num;
}

void RemoveUnderScore(char *str, char *outstr)
{
    while (*str)
    {
        if (*str != '_') *outstr++ = *str;
        str++;
    }
    *outstr = 0;
}

void StringConstant(char *str, char *outstr)
{
    int val = *str;

    if (val == '$')
    {
        str[0] = '0';
        str[1] = 'x';
        RemoveUnderScore(str+1, outstr+2);
    }
    else if (val == '%')
    {
	strcpy(outstr, ProcessBinaryConstant(str));
    }
    else if (val == '"')
    {
	int len = strlen(str) - 2;
	str++;
        *outstr = 0;
	while (len-- > 0)
	{
            *outstr++ = '\'';
            *outstr++ = *str++;
	    if (len > 0)
            {
                *outstr++ = ',';
                *outstr++ = ' ';
            }
	}
        *outstr = 0;
    }
    else if (isdigit(*str))
        RemoveUnderScore(str, outstr);
    else
        strcpy(outstr, str);
}

int PrintConstant(char *str, FILE *outfile)
{
    int num = 0;
    int val = *str;

    if (val == '$')
    {
        fprintf(outfile, "0x");
        num = PrintNumber(str+1, outfile) + 2;
    }
    else if (val == '%')
    {
        char *ptr = ProcessBinaryConstant(str);
	fprintf(outfile, "%s", ptr);
        num += strlen(ptr);
    }
    else if (val == '"')
    {
	int len = strlen(str) - 2;
	str++;
	while (len-- > 0)
	{
            num += 3;
	    fprintf(outfile, "'%c'", *str++);
	    if (len > 0)
            {
                num += 2;
                fprintf(outfile, ", ");
            }
	}
    }
    else if (isdigit(*str))
        num += PrintNumber(str, outfile);
    else
    {
        num += strlen(str);
        fprintf(outfile, "%s", str);
    }
    return num;
}

void ProcessVar(char **tokens, int numtokens, char *fname, FILE *outfile)
{
    int i, j;
    int type;
    int symnum;
    int mode;

    fprintf(outfile, "    %s", tokens[0]);

    type = SearchList(tokens[0], typelist);
    if (type < 0) type = 0;

    mode = 1;

    for (i = 1; i < numtokens; i++)
    {
	if (tokens[i][0] == '\'')
	{
	    fprintf(outfile, "; //%s\n", tokens[i] + 1);
	    return;
	}
        j = FindSymbol(tokens[i]);
        if (fname && j >= 0 && symbol[j].type == TYPE_CONSTANT &&
          symbol[j].array)
	    fprintf(outfile, " %s_%s", fname, tokens[i]);
        else
	    fprintf(outfile, " %s", tokens[i]);
	if (mode)
	{
	    symnum = AddSymbol(tokens[i], 0, TYPE_VAR);
	    //printf("ProcessVar: Add %s\n", tokens[i]);
	    if (i < numtokens - 1 && tokens[i+1][0] == '[')
	    {
	        //printf("ProcessVar: symnum = %d\n", symnum);
	        //printf("ProcessVar: Set array\n");
	        SetSymbolArray(symnum, 2);
		//printf("ProcessVar: array = %d\n", GetSymbolArray(symnum));
	    }
	    mode = 0;
	}
	if (tokens[i][0] == ',') mode = 1;
    }
    fprintf(outfile, ";\n");
}

void CheckDatState(DatStateT *state, int terminate, FILE *outfile)
{
    if (!state->mode) return;

    if (terminate)
    {
        if (state->mode == 1)
            fprintf(outfile, ";\n");
        else
            fprintf(outfile, "};\n");
        state->col = 0;
        state->mode = 0;
        state->notfirst = 0;
    }
    else if (state->notfirst)
    {
        if (state->col > 70)
        {
            state->col = 4;
            fprintf(outfile, ",\n    ");
        }
        else
        {
            state->col += 2;
            fprintf(outfile, ", ");
        }
    }
    else
    {
        if (state->col == 0)
        {
            state->col += 4;
            fprintf(outfile, "    ");
        }
        state->notfirst = 1;
    }
}

void ProcessDat(char **tokens, int numtokens, DatStateT *state, FILE *outfile)
{
    int i, j;

    if (numtokens == 0) return;

    for (i = 0; i < numtokens; i++)
    {
	j = FindSymbol(tokens[i]);
	if (GetSymbolType(j) == TYPE_DAT)
	{
            CheckDatState(state, 1, outfile);
#ifdef DAT_STRUCT
            fprintf("// %s\n", tokens[i]);
#else
            int size  = symbol[j].size;
            int array = symbol[j].array;
            if (array > 1)
            {
                fprintf(outfile, "static %s %s[%d] = {", types[size], symbol[j].name, array);
                state->col = strlen(symbol[j].name) + 20;
                state->mode = 2;
            }
            else
            {
                fprintf(outfile, "static %s %s = ", types[size], symbol[j].name);
                state->col = strlen(symbol[j].name) + 15;
                state->mode = 1;
            }
            state->notfirst = 0;
#endif
	}
	else if (IsConstant(tokens[i]))
	{
            //if (first) fprintf(outfile, "    ");
            CheckDatState(state, 0, outfile);
	    state->col += PrintConstant(tokens[i], outfile);
            //fprintf(outfile, ",");
	}
        else if (tokens[i][0] == '\'')
        {
            CheckDatState(state, 0, outfile);
            fprintf(outfile, " //%s\n", tokens[i]+1);
            state->col = 0;
            state->notfirst = 0;
        }
    }
    //if (!first) fprintf(outfile, "\n");
}

void PrintHexToken(char *str, FILE *outfile)
{
    fprintf(outfile, " 0x");
    str++;
    while (*str)
    {
	if (*str != '_') fputc(*str, outfile);
	str++;
    }
}

void PrintDecToken(char *str, FILE *outfile)
{
    fputc(' ', outfile);
    while (*str)
    {
	if (*str != '_') fputc(*str, outfile);
	str++;
    }
}

int EvaluateConstant(char **tokens, int numtokens, int index)
{
    for(;index < numtokens; index++)
    {
    }
}

void ProcessConstant(char **tokens, int numtokens, char *fname, FILE *outfile)
{
    int i, j;
    int parenflag;

    if (numtokens < 1)
    {
	fprintf(outfile, "\n");
	return;
    }

    parenflag = numtokens > 4 || (numtokens == 4 && tokens[3][0] != '\'');

    if (fname)
        fprintf(outfile, "#define %s_%s", fname, tokens[0]);
    else
        fprintf(outfile, "#define %s", tokens[0]);
    if (parenflag) fprintf(outfile, " (");
    for (i = 2; i < numtokens; i++)
    {
	if (tokens[i][0] == '$') PrintHexToken(tokens[i], outfile);
	else if(isdigit(tokens[i][0])) PrintDecToken(tokens[i], outfile);
	else if(tokens[i][0] == '\'')
	{
	    if (parenflag) fprintf(outfile, ")");
	    fprintf(outfile, " //%s\n", tokens[i] + 1);
	    return;
	}
	else
        {
            j = FindSymbol(tokens[i]);
            if (fname && j >= 0 && symbol[j].type == TYPE_CONSTANT &&
                symbol[j].array)
                fprintf(outfile, " %s_%s", fname, tokens[i]);
            else
                fprintf(outfile, " %s", tokens[i]);
        }
    }

    if (parenflag) fprintf(outfile, ")");
    fprintf(outfile, "\n");
}

char *PrintTokensToString(char **tokens, int tokenum)
{
    int i;
    int nospace = 1;
    char buffer[MAX_LINE_SIZE];
    char *ptr = buffer;
    int lastchar = 0;

    for (i = 0; i < tokenum; i++)
    {
        if (tokens[i][0] == '(' && nospace == 0 &&
            !isdigit(lastchar) && !isalpha(lastchar))
        {
            *ptr++ = ' ';
            strcpy(ptr, tokens[i]);
            ptr += strlen(tokens[i]);
        }
	else if (nospace || tokens[i][0] == ';' ||
	    tokens[i][0] == ',' ||
	    tokens[i][0] == '(' || tokens[i][0] == ')' ||
	    tokens[i][0] == '[' || tokens[i][0] == ']')
        {
            strcpy(ptr, tokens[i]);
            ptr += strlen(tokens[i]);
        }
	else
        {
            *ptr++ = ' ';
            strcpy(ptr, tokens[i]);
            ptr += strlen(tokens[i]);
        }
	if (strcmp(tokens[i], "0x") == 0 || strcmp(tokens[i], "(") == 0 ||
	    tokens[i][0] == '[' ||
	    tokens[i][0] == '&' ||
	    tokens[i][0] == '(' ||
	    strcmp(tokens[i], "//") ==  0)
	    nospace = 1;
	else
	    nospace = 0;
        lastchar = tokens[i][strlen(tokens[i]-1)];
    }
    *ptr = 0;
    ptr = malloc(strlen(buffer)+1);
    strcpy(ptr, buffer);
    return ptr;
}

void PrintTokens(char **tokens, int tokenum, FILE *outfile)
{
    int i;
    int nospace = 1;
    int lastchar = 0;

    PrintBlanks(indent, outfile);
    for (i = 0; i < tokenum; i++)
    {
        if (tokens[i][0] == '(' && nospace == 0 &&
            !isdigit(lastchar) && !isalpha(lastchar))
	    fprintf(outfile, " %s", tokens[i]);
	else if (nospace || tokens[i][0] == ';' ||
	    tokens[i][0] == ',' ||
	    tokens[i][0] == '(' || tokens[i][0] == ')' ||
	    tokens[i][0] == '[' || tokens[i][0] == ']')
	    fprintf(outfile, "%s", tokens[i]);
	else
	    fprintf(outfile, " %s", tokens[i]);
	if (strcmp(tokens[i], "0x") == 0 || strcmp(tokens[i], "(") == 0 ||
	    tokens[i][0] == '[' ||
	    tokens[i][0] == '&' ||
	    tokens[i][0] == '(' ||
	    strcmp(tokens[i], "//") ==  0)
	    nospace = 1;
	else
	    nospace = 0;
        lastchar = tokens[i][strlen(tokens[i]-1)];
    }
    fprintf(outfile, "\n");
}

int CheckComment(char *str, int *pmode, int pass)
{
    char *ptr = SkipChars(str, " \t");

    if (*pmode)
    {
        if (ptr[0] == '}')
	{
	    *pmode = 0;
	    if (pass) printf("*/\n");
            return 3;
	}
	else if (pass) printf("%s\n", str);
	return 1;
    }
    if (ptr[0] == '{')
    {
	if (pass) printf("/*\n");
        *pmode = 1;
	return 2;
    }
    return 0;
}

int CheckForUnicode(FILE *infile)
{
    int num;
    int unicode = 0;
    unsigned char buffer[2];

    num = fread(buffer, 1, 2, infile);

    if (num == 2 && buffer[0] == 0xff && buffer[1] == 0xfe) unicode = 1;

    fseek(infile, 0, SEEK_SET);

    return unicode;
}

int ReadString(char *buf, int size, FILE *infile, int unicode)
{
    int num;
    int count = 0;
    char *buf0 = buf;
    unsigned char bytepair[2];

    if (!unicode) count = (int)fgets(buf, size, infile);
    else
    {
        // Read unicode line
        while (count < size - 1)
        {
            num = fread(bytepair, 1, 2, infile);
            if (num != 2) break;
            if (bytepair[1] == 0)
                *buf++ = bytepair[0];
            else if (bytepair[0] == 0xff && bytepair[1] == 0xfe)
                continue;
            else if (bytepair[0] == 0x00 && bytepair[1] == 0x25)
                *buf++ = '-';
            else if (bytepair[0] == 0x02 && bytepair[1] == 0x25)
                *buf++ = '|';
            else
                *buf++ = '+';
            count++;
            if (buf[-1] == '\n') break;
        }
        *buf = 0;
    }
    if (!count)
        fseek(infile, 0, SEEK_SET);
    else
	RemoveCRLF(buf0);
    return count;
}

void ProcessIncludeFile(char *fname)
{
    char fname2[100];
    FILE *infile;
    char buffer[MAX_LINE_SIZE];
    char *ptr1;
    char *ptr2;
    int unicode;

    //printf("ProcessIncludeFile: %s\n", fname);
    strcpy(fname2, fname);
    strcat(fname2, ".h");
    infile = fopen(fname2, "r");

    if (infile == 0) return;

    unicode = CheckForUnicode(infile);

    while (ReadString(buffer, MAX_LINE_SIZE, infile, unicode))
    {
	//RemoveCRLF(buffer);
	//printf("trace 0 - %s\n", buffer);
	ptr1 = SkipChars(buffer, " \t");
	ptr2 = FindChars(ptr1, " \t");
	if (*ptr2 == 0) continue;
	//printf("trace 1\n");
	*ptr2 = 0;
	if (StrCompNoCase(ptr1, "long") != 0) continue;
	ptr1 = SkipChars(ptr2 + 1, " \t");
	ptr2 = FindChars(ptr1, " \t(");
	if (*ptr2 == 0) continue;
	*ptr2 = 0;
        AddSymbol(ptr1, 0, TYPE_PUB);
	//printf("//ProcessIncludeFile: %s\n", ptr1);
    }
    fclose(infile);
}

int UpdateMode(char *token0, int mode)
{
    if (StrCompNoCase(token0, "con") == 0) mode = MODE_CON;
    else if (StrCompNoCase(token0, "dat") == 0) mode = MODE_DAT;
    else if (StrCompNoCase(token0, "var") == 0) mode = MODE_VAR;
    else if (StrCompNoCase(token0, "obj") == 0) mode = MODE_OBJ;
    else if (StrCompNoCase(token0, "pub") == 0 ||
        StrCompNoCase(token0, "pri") == 0) mode = MODE_PUB;
    return mode;
}

void ProcessObjects(FILE *infile, int unicode)
{
    int num;
    char tokstr[MAX_LINE_SIZE*2];
    char buffer[MAX_LINE_SIZE];
    char *tokens[MAX_LINE_SIZE];
    int commentmode = 0;
    int mode = MODE_CON;

    while (ReadString(buffer, MAX_LINE_SIZE, infile, unicode))
    {
	if (CheckComment(buffer, &commentmode, 0)) continue;
	num = tokenize(tokens, buffer, tokstr);
	if (num == 0 || tokens[0][0] == '\'') continue;
        mode = UpdateMode(tokens[0], mode);
        if (mode == MODE_OBJ && num >= 3)
        {
	    int len = strlen(tokens[2]);
	    if (len > 1) tokens[2][len-1] = 0;
	    if (len > 2) tokens[2]++;
            AddSymbol(tokens[0], (int)CopyString(tokens[2]), TYPE_OBJECT);
	    //ProcessIncludeFile(tokens[2]);
        }
    }
}

void GenerateIncludes(FILE *infile, int unicode, FILE *outfile)
{
    int i;

    fprintf(outfile, "// INCLUDE FILES\n");
    fprintf(outfile, "#include \"spin.h\"\n");

    for (i = 0; i < numsym; i++)
    {
        if (symbol[i].type != TYPE_OBJECT) continue;
        fprintf(outfile, "#include \"%s.h\"\n", symbol[i].value);
    }

    fprintf(outfile, "\n");
}

void ProcessConstants(FILE *infile, int unicode, char *fname, FILE *outfile)
{
    int j;
    int num;
    char tokstr[MAX_LINE_SIZE*2];
    char buffer[MAX_LINE_SIZE];
    char *tokens[MAX_LINE_SIZE];
    int commentmode = 0;
    int mode = MODE_CON;

    fprintf(outfile, "// CONSTANTS\n");

    while (ReadString(buffer, MAX_LINE_SIZE, infile, unicode))
    {
	if (CheckComment(buffer, &commentmode, 0)) continue;
	num = tokenize(tokens, buffer, tokstr);
	if (num == 0 || tokens[0][0] == '\'') continue;
        mode = UpdateMode(tokens[0], mode);
        if (mode == MODE_CON && num >= 3)
        {
            if (fname)
            {
                j = AddSymbol(tokens[0], 0, TYPE_CONSTANT);
                symbol[j].array = 1;
            }
            ProcessConstant(tokens, num, fname, outfile);
        }
    }

    fprintf(outfile, "\n");
}

void ProcessVars(FILE *infile, int unicode, char *fname, FILE *outfile, int header)
{
    int i;
    int num;
    char tokstr[MAX_LINE_SIZE*2];
    char buffer[MAX_LINE_SIZE];
    char *tokens[MAX_LINE_SIZE];
    int commentmode = 0;
    int mode = MODE_CON;

    fprintf(outfile, "// CLASS DECLARATION\n");
    fprintf(outfile, "class %s_Class {\n", fname);
    fprintf(outfile, "private:\n");
    if (!header) fname = 0;

    while (ReadString(buffer, MAX_LINE_SIZE, infile, unicode))
    {
	if (CheckComment(buffer, &commentmode, 0)) continue;
	num = tokenize(tokens, buffer, tokstr);
	if (num == 0 || tokens[0][0] == '\'') continue;
        mode = UpdateMode(tokens[0], mode);
        if (mode == MODE_VAR && num >= 2)
        {
            ProcessVar(tokens, num, fname, outfile);
        }
    }

    for (i = 0; i < numsym; i++)
    {
        if (symbol[i].type != TYPE_OBJECT) continue;
        fprintf(outfile, "    %s_Class %s;\n", symbol[i].value, symbol[i].name);
    }

    ProcessPrototypes(infile, unicode, "pri", outfile);
    fprintf(outfile, "public:\n");
    ProcessPrototypes(infile, unicode, "pub", outfile);
    fprintf(outfile, "};\n");
    fprintf(outfile, "\n");
}

void ProcessDats(FILE *infile, int unicode, FILE *outfile)
{
    int i, j;
    int num;
    int lastsym = -1;
    char tokstr[MAX_LINE_SIZE*2];
    char buffer[MAX_LINE_SIZE];
    char *tokens[MAX_LINE_SIZE];
    int commentmode = 0;
    int mode = MODE_CON;
    DatStateT state = {0, 0, 0};

    fprintf(outfile, "// DAT VARIABLES\n");

    // Add DAT labels to symbol table and determine the size of each
    while (ReadString(buffer, MAX_LINE_SIZE, infile, unicode))
    {
	if (CheckComment(buffer, &commentmode, 0)) continue;
	num = tokenize(tokens, buffer, tokstr);
	if (num == 0 || tokens[0][0] == '\'') continue;
        mode = UpdateMode(tokens[0], mode);
        if (mode == MODE_DAT && StrCompNoCase(tokens[0], "dat"))
        {
            j = SearchList(tokens[0], typelist);
            if (j < 0)
            {
                lastsym = AddSymbol(tokens[0], 0, TYPE_DAT);
                if (num < 2) continue;
                j = SearchList(tokens[1], typelist);
                if (j < 0) continue;
            }
            SetSymbolSize(lastsym, j);
            IncrSymbolArray(lastsym);
            // Count commas
            for (i = 1; i < num; i++)
            {
                if (strcmp(tokens[i], ",") == 0)
                    IncrSymbolArray(lastsym);
            }
        }
    }

#ifdef DAT_STRUCT
    // Generate the DAT struct
    fprintf(outfile, "typedef struct DatS {\n");
    for (i = 0; i < numsym; i++)
    {
        int type, array, size;

        type  = symbol[i].type;
        if (type != TYPE_DAT) continue;

        size  = symbol[i].size;
        array = symbol[i].array;
        if (array > 1)
            fprintf(outfile, "    %s %s[%d];\n", types[size], symbol[i].name, array);
        else
            fprintf(outfile, "    %s %s;\n", types[size], symbol[i].name);
    }
    fprintf(outfile, "} DatT;\n");
    fprintf(outfile, "\n");
    fprintf(outfile, "DatT dat = {\n");
#endif

    // Initialize the DAT struct
    while (ReadString(buffer, MAX_LINE_SIZE, infile, unicode))
    {
	if (CheckComment(buffer, &commentmode, 0)) continue;
	num = tokenize(tokens, buffer, tokstr);
	if (num == 0 || tokens[0][0] == '\'') continue;
        mode = UpdateMode(tokens[0], mode);
        if (mode == MODE_DAT && StrCompNoCase(tokens[0], "dat"))
        {
            ProcessDat(tokens, num, &state, outfile);
        }
    }

    CheckDatState(&state, 1, outfile);
#ifdef DAT_STRUCT
    fprintf(outfile, "};\n");
#endif
    fprintf(outfile, "\n");
}

void ProcessPrototypes(FILE *infile, int unicode, char *pub_pri, FILE *outfile)
{
    int i;
    int num;
    int first;
    int pubnum = 1;
    char tokstr[MAX_LINE_SIZE*2];
    char buffer[MAX_LINE_SIZE];
    char *tokens[MAX_LINE_SIZE];
    int commentmode = 0;

    while (ReadString(buffer, MAX_LINE_SIZE, infile, unicode))
    {
	if (CheckComment(buffer, &commentmode, 0)) continue;
	num = tokenize(tokens, buffer, tokstr);
	if (num == 0 || tokens[0][0] == '\'') continue;
	if (StrCompNoCase(tokens[0], pub_pri) == 0)
	{
            AddSymbol(tokens[1], pubnum++, TYPE_PUB);
	    fprintf(outfile, "    long %s(", tokens[1]);
	    if (num < 3 || tokens[2][0] != '(')
	    {
		fprintf(outfile, "void);\n");
		continue;
	    }
            first = 1;
	    for (i = 3; i < num; i++)
	    {
		if (tokens[i][0] == ')') break;
		if (tokens[i][0] == ',') fprintf(outfile, ",");
		else
                {
                    if (first)
                        first = 0;
                    else
                        fprintf(outfile, " ");
                    fprintf(outfile, "long %s", tokens[i]);
                }
	    }
	    fprintf(outfile, ");\n");
        }
    }
}

void ProcessPubs(FILE *infile, int unicode, char *fname, FILE *outfile)
{
    int i, j, num;
    char buffer[MAX_LINE_SIZE];
    char tokstr[MAX_LINE_SIZE*2];
    char *tokens[MAX_LINE_SIZE];
    char *tokens2[MAX_LINE_SIZE];
    int tokenum2 = 0;
    int mode = 0;
    int pubnum = 1;
    int type;
    int commentmode = 0;
    char *loopvar;
    int repeat_flag;
    int unaryflag;
    int trailing_comment;

    levind[0] = 0;
    levtype[0] = 0;

    commentmode = 0;
    mode = 1; // Default is CON
    while (ReadString(buffer, MAX_LINE_SIZE, infile, unicode))
    {
	tokenum2 = 0;
        loopvar = 0;
        repeat_flag = 0;
	if ((i = CheckComment(buffer, &commentmode, 0)))
        {
            if (i == 2)
                AddComment("/*");
            else if (i == 3)
                AddComment("*/");
            else
                AddComment(buffer);
            continue;
        }
	indent = GetIndent(buffer);
	num = tokenize(tokens, buffer, tokstr);
	if (num == 0)
	{
	    //printf(outfile, "\n");
	    continue;
	}
	if (tokens[0][0] == '\'')
	{
            for (i = 0; i < indent; i++) *comptr++ = ' ';
            *comptr++ = '/';
            *comptr++ = '/';
            if (tokens[0][1] == '\'')
	        AddComment(tokens[0]+2);
            else
	        AddComment(tokens[0]+1);
#if 0
	    PrintBlanks(indent, outfile);
            if (tokens[0][1] == '\'')
	        fprintf(outfile, "//%s\n", tokens[0]+2);
            else
	        fprintf(outfile, "//%s\n", tokens[0]+1);
#endif
	    continue;
	}

        // Check if indentation has changed
	if (indent > levind[level])
	{
	    PrintBlanks(levind[level], outfile);
	    if (mode == 4) fprintf(outfile, "{\n");
	    levind[++level] = indent;
	    levtype[level] = 0;
	}
	else if (indent < levind[level])
	{
	    while (indent < levind[level])
	    {
	        if (level == 1 && mode == 4 && levtype[level] != LEVEL_TYPE_RETURN)
		{
		    fprintf(outfile, "  return %s;\n", pubptr);
		}
		if (levtype[level - 1] == LEVEL_TYPE_CASE_ITEM)
		{
		    PrintBlanks(levind[level], outfile);
		    fprintf(outfile, "break;\n");
		}
		PrintBlanks(levind[--level], outfile);
	        if (mode == 4) fprintf(outfile, "}\n");
		if (levtype[level] == LEVEL_TYPE_REPEAT && StrCompNoCase(tokens[0], "while") != 0)
		{
		    PrintBlanks(levind[level], outfile);
		    fprintf(outfile, "while (1);\n");
		}
                else if (levtype[level] == LEVEL_TYPE_REPEAT_FROM)
                {
		    PrintBlanks(levind[level], outfile);
#if 0
		    fprintf(outfile, "while (repeat_from(x,y,z));\n");
#else
                    fprintf(outfile, "while(repeat_from(%s\n", levstr[level]);
                    //free(levstr[level]);
                    levtype[level] = 0;
#endif
                }
	    }
	}

        mode = UpdateMode(tokens[0], mode);
        if (mode != MODE_PUB) continue;

        // Check for the first line of a method
	if (StrCompNoCase(tokens[0], "pub") == 0 ||
	    StrCompNoCase(tokens[0], "pri") == 0)
	{
	    int locmode, locspace, locaddr;

            fprintf(outfile, "\n");
            PrintComments(outfile);
	    //mode = 4;
	    pubptr = "result";
            //PrintSymbolTable();
            RemoveLocalSymbols();
            locmode = locspace = 0;
            locaddr = 4;
	    PrintBlanks(indent, outfile);
	    fprintf(outfile, "long %s_Class::%s", fname, tokens[1]);
	    if (tokens[2][0] == '(')
	    {
		fprintf(outfile, "(");
		i = 3;
	    }
	    else
	    {
		fprintf(outfile, "(void)\n");
	        PrintBlanks(indent, outfile);
	        fprintf(outfile, "{\n");
	        levind[++level] = indent + 2;
                if (num <= 2 || tokens[2][0] != ':')
                {
	            PrintBlanks(indent + 2, outfile);
	            fprintf(outfile, "long result = 0;\n");
                }
	        //PrintBlanks(indent + 2, outfile);
		//printf(outfile, "long repeat_count;\n");
		i = 2;
	    }
            for (; i < num; i++)
            {
		if (tokens[i][0] == '(') continue;
		if (tokens[i][0] == ')')
		{
		    fprintf(outfile, ")\n");
		    fprintf(outfile, "{\n");
		    levind[++level] = indent + 2;
                    if (num <= i + 1 || tokens[i + 1][0] != ':')
                    {
	                PrintBlanks(indent + 2, outfile);
		        fprintf(outfile, "long result = 0;\n");
                    }
		    continue;
		}
		if (tokens[i][0] == ',')
		{
		    fprintf(outfile, ", ");
		    continue;
		}
                if (tokens[i][0] == 0) break;
                if (tokens[i][0] == '\'')
		{
		    fprintf(outfile, " %s", tokens[i]);
		    break;
		}
                if (tokens[i][0] == '|')
		{
	            PrintBlanks(indent+2, outfile);
		    fprintf(outfile, "long ");
		    locmode = 1;
		    continue;
		}
		else if (tokens[i][0] == ':' && i < num - 1)
		{
	            PrintBlanks(indent + 2, outfile);
		    fprintf(outfile, "long %s = 0;\n", tokens[++i]);
		    pubptr = bigptr;
	            AddSymbol(tokens[i], 0, TYPE_LOCAL);
		    continue;
		}
                else
                {
                    //printf(outfile, " ADD %s", tokens[i]);
	            AddSymbol(tokens[i], locaddr, TYPE_LOCAL);
	            locaddr += 4;
	            if (locmode) locspace += 4;
                }
		if (locmode)
		    fprintf(outfile, "%s", tokens[i]);
	        else
		    fprintf(outfile, " long %s", tokens[i]);
            }
	    if (locmode) fprintf(outfile, ";\n");
	}

        // Process a statement within a method
	else
	{
	    int bracketflag = 0;
	    int printsemi = 1;
	    int printparen = 0;

            PrintComments(outfile);
            levtype[level] = 0;
            i = 0;
	    if (StrCompNoCase(tokens[0], "repeat") == 0)
	    {
		if (num == 1)
		{
		    levtype[level] = LEVEL_TYPE_REPEAT;
#if 1
                    PrintBlanks(indent, outfile);
		    fprintf(outfile, "do\n");
		    continue;
#else
                    printsemi = 0;
		    tokens2[0] = "do";
                    tokenum2 = 1;
                    break;
#endif
		}
		else if (StrCompNoCase(tokens[1], "while") == 0)
		{
		    tokens2[tokenum2++] = "while ";
		    tokens2[tokenum2++] = "(";
		    printparen = 1;
		    printsemi = 0;
		    i = 2;
		}
		else if (StrCompNoCase(tokens[1], "until") == 0)
		{
		    tokens2[tokenum2++] = "while (!";
		    tokens2[tokenum2++] = "(";
		    printparen = 2;
		    printsemi = 0;
		    i = 2;
		}
		else if (num > 3 && StrCompNoCase(tokens[2], "from") == 0)
		{
                    int type = 0;

                    PrintBlanks(indent, outfile);
                    if (0)
                    {
                        repeat_flag = REPEAT_FOR;
		        tokens2[tokenum2++] = "for (";
		        printparen = 1;
		        printsemi = 0;
                    }
                    else
                    {
		        levtype[level] = LEVEL_TYPE_REPEAT_FROM;
                        levstr[level] = "REPEAT_FROM_NULL\n";
                        repeat_flag = REPEAT_FROM;
                        //printf("Setting repeat_flag, level = %d\n", level);
                    }
                    j = FindSymbol(tokens[1]);
                    if (j >= 0) type = symbol[j].type;
#ifdef DAT_STRUCT
                    if (GetSymbolType(j) == TYPE_DAT)
		        tokens2[tokenum2++] = "dat.";
#endif
                    if (j >= 0)
		        tokens2[tokenum2++] = symbol[j].name;
                    else
		        tokens2[tokenum2++] = tokens[1];
		    tokens2[tokenum2++] = "=";
                    loopvar = tokens[1];
		    i = 3;
		}
		else
		{
                    repeat_flag = REPEAT_COUNT;
		    tokens2[tokenum2++] = "for (int repeat_count = ";
		    i = 1;
		}
	    }

	    if (StrCompNoCase(tokens[0], "while") == 0)
	    {
	        tokens2[tokenum2++] = "while ";
	        tokens2[tokenum2++] = "(";
	        printparen = 1;
	        printsemi = 1;
	        i = 1;
	    }

	    if (level > 0 && levtype[level-1] == LEVEL_TYPE_CASE)
	    {
		levtype[level] = LEVEL_TYPE_CASE_ITEM;
		tokens2[tokenum2++] = "case";
		printsemi = 0;
	    }

	    if (StrCompNoCase(tokens[0], "if") == 0)
	    {
		tokens2[tokenum2++] = "if ";
		tokens2[tokenum2++] = "(";
		printparen = 1;
		printsemi = 0;
		i = 1;
	    }
	    else if (StrCompNoCase(tokens[0], "ifnot") == 0)
	    {
		tokens2[tokenum2++] = "if (!";
		tokens2[tokenum2++] = "(";
		printparen = 2;
		printsemi = 0;
		i = 1;
	    }
	    else if (StrCompNoCase(tokens[0], "else") == 0)
	    {
		PrintBlanks(indent, outfile);
		fprintf(outfile, "else\n");
		continue;
	    }
	    else if (StrCompNoCase(tokens[0], "elseif") == 0)
	    {
		tokens2[tokenum2++] = "else if (";
		printparen = 1;
		printsemi = 0;
		i = 1;
	    }
	    else if (StrCompNoCase(tokens[0], "case") == 0)
	    {
		levtype[level] = LEVEL_TYPE_CASE;
		tokens2[tokenum2++] = "switch (";
		printparen = 1;
		printsemi = 0;
		i = 1;
	    }
	    else if (StrCompNoCase(tokens[0], "abort") == 0)
	    {
		tokens2[tokenum2++] = "longjmp";
		tokens2[tokenum2++] = "(envbuf,";
		printparen = 1;
		printsemi = 1;
		i = 1;
	    }
	    else if (StrCompNoCase(tokens[0], "return") == 0)
	    {
		levtype[level] = LEVEL_TYPE_RETURN;
		tokens2[tokenum2++] = "return";
		if (num < 2 || tokens[1][0] == '\'')
		{
		    tokens2[tokenum2++] = pubptr;
		}
		printsemi = 1;
		i = 1;
	    }
            unaryflag = 0;
            trailing_comment = 0;
            for (; i < num; i++)
	    {
                //printf("Loop %d: tokens[i] = %s\n", i, tokens[i]);
		j = SearchOpsList(tokens[i], OpsList);
		if (j >= 0)
		{
                    if (i == 0 &&
                        strcmp(tokens[0], "++") && strcmp(tokens[0], "--"))
                    {
                        unaryflag = j + 1;
                        continue;
                    }
		    if (OpsList[j].flag == 0)
		    {
			tokens2[tokenum2++] = tokens[i];
			continue;
		    }
		    else if (OpsList[j].flag == 1)
		    {
			tokens2[tokenum2++] = OpsList[j].cstr;
			continue;
		    }
		    else if (OpsList[j].flag == 2)
		    {
		        int ii = LocatePrevVar(tokens2, tokenum2, tokenum2-1);
		        //int ii = LocatePrevOp(tokens2, tokenum2, j) + 1;
		        tokenum2 = InsertToken(tokens2, tokenum2, ii, "(");
		        tokenum2 = InsertToken(tokens2, tokenum2, ii, OpsList[j].cstr);
		        tokens2[tokenum2++] = ",";
		        ii = LocateNextVar(tokens, num, i + 1);
		        num = InsertToken(tokens, num, ii, ")");
			continue;
		    }
		    else if (OpsList[j].flag == 3)
		    {
		        if (i < num - 1 && tokens[i+1][0] != '\'')
			{
		            int ii;
		            tokens2[tokenum2++] = OpsList[j].cstr;
		            tokens2[tokenum2++] = "(";
		            ii = LocateNextVar(tokens, num, i + 1);
		            num = InsertToken(tokens, num, ii, ")");
			    continue;
			}
		    }
		    else if (OpsList[j].flag == 5)
		    {
		        if (i < num - 1 && tokens[i+1][0] != '\'')
			{
		            int ii;
		            tokens2[tokenum2++] = "=";
		            tokens2[tokenum2++] = OpsList[j].cstr;
		            tokens2[tokenum2++] = "(";
		            tokens2[tokenum2++] = tokens2[tokenum2-4];
		            tokens2[tokenum2++] = ",";
                            printparen = 1;
                            if (i + 2 < num && tokens[i+2][0] != '\'')
                            {
		                tokens2[tokenum2++] = "(";
                                printparen++;
                            }
		            //ii = LocateNextVar(tokens, num, i + 1);
		            //num = InsertToken(tokens, num, ii, ")");
			    continue;
			}
		    }
		    else if (OpsList[j].flag == 6)
		    {
		        if (i < num - 1 && tokens[i+1][0] != '\'')
			{
		            tokens2[tokenum2++] = "=";
		            tokens2[tokenum2++] = tokens2[tokenum2-2];
		            tokens2[tokenum2++] = OpsList[j].cstr;
                            if (i + 2 < num && tokens[i+2][0] != '\'')
                            {
		                tokens2[tokenum2++] = "(";
                                printparen = 1;
                            }
			    continue;
			}
		    }
		}

		if (tokens[i][0] == '$')
		{
		    tokens2[tokenum2++] = "0x";
                    RemoveUnderScore(tokens[i]+1, tokens[i]);
		    tokens2[tokenum2++] = tokens[i];
		}
		else if (tokens[i][0] == '%')
		{
		    tokens2[tokenum2++] = ProcessBinaryConstant(tokens[i]);
		}
		else if (tokens[i][0] == '\\')
		{
		    if (tokenum2 >= 2 && tokens2[tokenum2-1][0] == '=')
		    {
			int ii = tokenum2 - 2;
			tokenum2 = InsertToken(tokens2, tokenum2, ii++, "if (!(");
			tokenum2 = InsertToken(tokens2, tokenum2, ii, tokens2[ii]);
			ii++;
			tokenum2 = InsertToken(tokens2, tokenum2, ii, tokens2[ii+1]);
			ii++;
			tokenum2 = InsertToken(tokens2, tokenum2, ii, "setjmp(envbuf)))");
		    }
		    else
		    {
		        tokens2[tokenum2++] = "if (!setjmp(envbuf))";
		    }
		    continue;
		}
		else if (strcmp(tokens[i], "~") == 0)
		{
		    tokens2[tokenum2++] = "=";
		    tokens2[tokenum2++] = "0";
		}
		else if (strcmp(tokens[i], "~~") == 0)
		{
		    tokens2[tokenum2++] = "=";
		    tokens2[tokenum2++] = "-1";
		}
		else if (tokens[i][0] == '"' && tokens[i][2] == '"')
		{
		    tokens[i][0] = '\'';
		    tokens[i][2] = '\'';
		    tokens2[tokenum2++] = tokens[i];
		}
		else if (tokens[i][0] == '@')
		{
		    tokens2[tokenum2++] = "(long)";
		    if (i < num - 1 && (j = FindSymbol(tokens[i + 1])) >= 0)
		    {
	                //printf(outfile, "Symbol number = %d\n", j);
			//printf(outfile, "\nSymbol = %s, array = %d\n",
			    //tokens[i+1], GetSymbolArray(j));
			if (GetSymbolArray(j) <= 1 ||
                            (i < num - 2 && tokens[i+2][0] == '['))
			{
		            tokens2[tokenum2++] = "&";
			}
		    }
		    else
		    {
		        tokens2[tokenum2++] = tokens[i];
		    }
		    tokens2[tokenum2++] = tokens[++i];
		}
		else if (StrCompNoCase(tokens[i], "byte") == 0)
		{
		    if (i < num - 1 && tokens[i+1][0] == '[')
		    {
			tokens2[tokenum2++] = "BYTE";
			tokens2[tokenum2++] = "(";
			i++;
			bracketflag = 1;
		    }
		    else
		    {
		        tokens2[tokenum2++] = tokens[i];
		    }
		}
		else if (StrCompNoCase(tokens[i], ".byte") == 0)
		{
		    if (tokenum2 > 0)
		    {
			char *ptr = tokens2[--tokenum2];
			tokens2[tokenum2++] = "BYTE";
			tokens2[tokenum2++] = "(";
			tokens2[tokenum2++] = "&";
			tokens2[tokenum2++] = ptr;
			tokens2[tokenum2++] = ")";
		    }
		}
		else if (StrCompNoCase(tokens[i], "word") == 0)
		{
		    if (i < num - 1 && tokens[i+1][0] == '[')
		    {
			tokens2[tokenum2++] = "WORD";
			tokens2[tokenum2++] = "(";
			i++;
			bracketflag = 1;
		    }
		    else
		    {
		        tokens2[tokenum2++] = tokens[i];
		    }
		}
		else if (StrCompNoCase(tokens[i], ".word") == 0)
		{
		    if (tokenum2 > 0)
		    {
			char *ptr = tokens2[--tokenum2];
			tokens2[tokenum2++] = "WORD";
			tokens2[tokenum2++] = "(";
			tokens2[tokenum2++] = "&";
			tokens2[tokenum2++] = ptr;
			tokens2[tokenum2++] = ")";
		    }
		}
		else if (StrCompNoCase(tokens[i], "long") == 0)
		{
		    if (i < num - 1 && tokens[i+1][0] == '[')
		    {
			tokens2[tokenum2++] = "LONG";
			tokens2[tokenum2++] = "(";
			i++;
			bracketflag = 1;
		    }
		    else
		    {
		        tokens2[tokenum2++] = tokens[i];
		    }
		}
		else if (tokens[i][0] == ']' && bracketflag)
		{
		    if (i < num -1 && tokens[i+1][0] == '[')
		    {
		        tokens2[tokenum2++] = ")";
		    }
		    else
		    {
		        tokens2[tokenum2++] = ")[0]";
		    }
		    bracketflag = 0;
		}
		else if (StrCompNoCase(tokens[i], "constant") == 0)
		{
		    continue;
		}
		else if (StrCompNoCase(tokens[i], "to") == 0)
		{
		    tokens2[tokenum2++] = ";";
                    if (repeat_flag == REPEAT_FOR)
                    {
                        if (loopvar)
		            tokens2[tokenum2++] = loopvar;
                        else
		            tokens2[tokenum2++] = "NULL";
		        tokens2[tokenum2++] = "<=";
                    }
                    else
                    {
                        PrintTokens(tokens2, tokenum2, outfile);
                        PrintBlanks(indent, outfile);
                        fprintf(outfile, "do\n");
                        tokenum2--;
		        tokens2[1] = ",";
		        tokens2[tokenum2++] = ",";
                    }
                    continue;
		}
		else if (StrCompNoCase(tokens[i], "step") == 0)
		{
                    if (repeat_flag == REPEAT_FOR)
                    {
		        tokens2[tokenum2++] = ";";
                        if (loopvar)
		            tokens2[tokenum2++] = loopvar;
                        else
		            tokens2[tokenum2++] = "NULL";
		        tokens2[tokenum2++] = "+=";
                        loopvar = 0;
                    }
                    else
                    {
		        tokens2[tokenum2++] = ",";
                        repeat_flag = REPEAT_STEP;
                    }
		    continue;
		}
		else if (StrCompNoCase(tokens[i], "string") == 0)
		{
		    int parencount = 1;
		    char *ptr;
		    int j;

		    if (tokens[++i][0] != '(') continue;
		    i++;
		    tokens2[tokenum2++] = "(long)";
		    tokens2[tokenum2++] = bigptr;
		    ptr = bigptr;
		    *ptr++ = '"';
		    for (;i < num; i++)
		    {
			if (tokens[i][0] == '(') parencount++;
			else if (tokens[i][0] == ')') parencount--;
			else if (tokens[i][0] == ',') continue;
			else if (tokens[i][0] == '"')
			{
			    strcpy(ptr, tokens[i] + 1);
			    ptr += strlen(ptr) - 1;
			}
			else if (tokens[i][0] >= '0' && tokens[i][0] <= '9')
			{
			    sscanf(tokens[i], "%d", &j);
			    sprintf(ptr, "\\x%2.2x", j&255);
			    ptr += strlen(ptr);
			}
			else if (tokens[i][0] == '$')
			{
			    sscanf(tokens[i]+1, "%x", &j);
			    sprintf(ptr, "\\x%2.2x", j&255);
			    ptr += strlen(ptr);
			}
			if (parencount <= 0) break;
		    }
		    *ptr++ = '"';
		    *ptr = 0;

		    bigptr += strlen(bigptr) + 1;
		}
		else if (tokens[i][0] == '\'')
		{
                    //printf("Set trailing_comment\n");
                    trailing_comment = 1;
                    break;
#if 0
		    while (printparen)
		    {
                        if (printparen == 1 && loopvar)
                        {
		            tokens2[tokenum2++] = ";";
		            tokens2[tokenum2++] = loopvar;
		            tokens2[tokenum2++] = "++";
                        }
			tokens2[tokenum2++] = ")";
			printparen--;
		    }
		    if (printsemi)
		    {
			tokens2[tokenum2++] = ";";
			printsemi = 0;
		    }
		    tokens2[tokenum2++] = "//";
		    tokens2[tokenum2++] = tokens[i] + 1;
#endif
		}
		else
		{
                    int type = 0;
		    char *ptr = FindChar(tokens[i], '.');
		    //if (*ptr) *ptr = '_';
		    j = FindSymbol(tokens[i]);
                    type = GetSymbolType(j);
#ifdef DAT_STRUCT
                    if (GetSymbolType(j) == TYPE_DAT)
                        tokens2[tokenum2++] = "dat.";
#endif
                    if (j >= 0)
		        tokens2[tokenum2++] = symbol[j].name;
                    else if (IsConstant(tokens[i]))
                    {
                        char tempstr[100];
                        StringConstant(tokens[i], tempstr);
		        tokens2[tokenum2++] = CopyString(tempstr);
                    }
                    else
		        tokens2[tokenum2++] = tokens[i];
		    //Check for function with no parms
                    if ((*ptr || (j >= 0 && GetSymbolType(j) == TYPE_PUB)) &&
		        (i >= num - 1 || tokens[i+1][0] != '('))
		        tokens2[tokenum2++] = "()";
                    if (type == TYPE_VAR || type == TYPE_DAT)
                    {
                        if (symbol[j].array > 1 && i+1 < num && tokens[i+1][0] != '[')
		            tokens2[tokenum2++] = "[0]";
                        
                    }
		}
	    }

            //printf("trailing_comment = %d, printparen = %d, printsemi = %d\n",
                //trailing_comment, printparen, printsemi);

            if (repeat_flag == REPEAT_COUNT)
            {
	        tokens2[tokenum2++] = ";";
	        tokens2[tokenum2++] = "repeat_count > 0; repeat_count--)";
                PrintTokens(tokens2, tokenum2, outfile);
                continue;
            }
            else if (repeat_flag)
            {
                if (repeat_flag != REPEAT_STEP)
                {
	            tokens2[tokenum2++] = ",";
	            tokens2[tokenum2++] = "1";
                }
	        tokens2[tokenum2++] = "));";
                levstr[level] = PrintTokensToString(tokens2, tokenum2);
                continue;
            }

	    while (printparen)
	    {
                if (printparen == 1 && loopvar)
                {
	            tokens2[tokenum2++] = ";";
	            tokens2[tokenum2++] = loopvar;
	            tokens2[tokenum2++] = "++";
                }
		tokens2[tokenum2++] = ")";
		printparen--;
	    }

            if (unaryflag)
            {
                int num = tokenum2;
		tokens2[tokenum2++] = "=";
                j = unaryflag - 1;
		if (OpsList[j].flag == 3)
                {
	            tokens2[tokenum2++] = OpsList[j].cstr;
		    tokens2[tokenum2++] = "(";
                    printparen = 1;
                }
		else if (OpsList[j].flag == 0)
                {
		    tokens2[tokenum2++] = tokens[0];
                }
		else
                {
	            tokens2[tokenum2++] = OpsList[j].cstr;
                }
                for (i = 0; i < num; i++)
                    tokens2[tokenum2++] = tokens2[i];
                if (printparen)
		    tokens2[tokenum2++] = ")";
            }
 
	    if (printsemi) tokens2[tokenum2++] = ";";

            if (trailing_comment)
            {
                tokens2[tokenum2++] = "//";
                tokens2[tokenum2++] = tokens[num-1] + 1;
            }
           
            PrintTokens(tokens2, tokenum2, outfile);
	}
    }
    //PrintSymbolTable();
    while (level > 0)
    {
	if (level == 1 && mode == 4) fprintf(outfile, "  return %s;\n", pubptr);
	PrintBlanks(levind[--level], outfile);
	fprintf(outfile, "}\n");
	if (levtype[level] == LEVEL_TYPE_REPEAT && StrCompNoCase(tokens[0], "while") != 0)
	{
	    PrintBlanks(levind[level], outfile);
	    fprintf(outfile, "while (1);\n");
	}
    }
    PrintComments(outfile);
}

FILE *OpenFile(char *fname, char *mode)
{
    FILE *stream = fopen(fname, mode);

    if (stream == 0)
    {
	printf("Can't open %s\n", fname);
	exit(1);
    }

    return stream;
}

int main(int argc, char **argv)
{
    char *extptr;
    int unicode;
    FILE *infile;
    FILE *outfile;
    char fname[80];

    comptr = combuf;
    *comptr = 0;

    if (argc != 2) usage();

    strcpy(fname, argv[1]);
    extptr = FindChar(fname, '.');
    if (*extptr == 0) strcpy(extptr, ".spin");
    infile = OpenFile(fname, "r");
    *extptr = 0;

    unicode = CheckForUnicode(infile);

    ProcessObjects(infile, unicode);

    // Generate Header File
    strcpy(extptr, ".h");
    outfile = OpenFile(fname, "w");
    *extptr = 0;
    ProcessConstants(infile, unicode, fname, outfile);
    ProcessVars(infile, unicode, fname, outfile, 1);
    fclose(outfile);

    // Generate CPP File
    strcpy(extptr, ".cpp");
    outfile = OpenFile(fname, "w");
    *extptr = 0;
    GenerateIncludes(infile, unicode, outfile);
    ProcessConstants(infile, unicode, 0, outfile);
    ProcessVars(infile, unicode, fname, outfile, 0);
    ProcessDats(infile, unicode, outfile);
    ProcessPubs(infile, unicode, fname, outfile);
    fclose(outfile);

    return 0;
}
